/*
 * Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, School of Computing, National University of Singapore, 2005.
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The library will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - No redistribution of the code is allowed
 *  - All bug fixes will be returned to zhangru1@comp.nus.edu.sg for
 *    inclusion in future releases
 *
 */

/*************************************************************************

*
* File : bt_io.c
*
* Modules:
* Routines to read and write content of INODE and LNODE from and to files
*   BT_IOReadLNode()
*   BT_IOReadINode()
*   BT_IOReadONode()
*   BT_IOWriteLNode()
*   BT_IOWriteINode()
*   BT_IOWriteONode()
*   BT_IOWriteNewLNode()
*   BT_IOWriteNewINode()
*   BT_IOWriteNewONode()
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/
#include "btree.h"
extern int max_top; 

/* function prototype */
/* internal */
int BT_IOReadLNode(BTREE *, PTR, LNODE *);
int BT_IOReadINode(BTREE *, PTR, INODE *);
int BT_IOReadONode(BTREE *, PTR, ONODE *);
int BT_IOWriteLNode(BTREE *, PTR, LNODE *);
int BT_IOWriteINode(BTREE *, PTR, INODE *);
int BT_IOWriteONode(BTREE *, PTR, ONODE *);
PTR BT_IOWriteNewLNode(BTREE *, LNODE *);
PTR BT_IOWriteNewINode(BTREE *, INODE *);
PTR BT_IOWriteNewONode(BTREE *, ONODE *);

/* external */
extern void  IO_PageAccess(FILE *, long);

/***************************************************************************

* BT_IOReadLNode():
*   Routine to read a LNODE from the b+ tree file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successsful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOReadLNode(BTREE *btree,
                   PTR offset, 
                   LNODE *node
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fread(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
      return result;
    } else {
      printf("BT_IOReadLNode : read node failure \n");
      return result;
    }
  } else {
    printf("BT_IOReadLNode : seek node failure \n");
    return result;
  }
}

/***************************************************************************

* BT_IOReadONode():
*   Routine to read a ONODE from the b+ tree file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successsful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOReadONode(BTREE *btree,
                   PTR offset, 
                   ONODE *node
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fread(node, O_PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
      return result;
    } else {
      printf("BT_IOReadONode : read node failure \n");
      return result;
    }
  } else {
    printf("BT_IOReadONode : seek node failure \n");
    return result;
  }
}

/***************************************************************************

* BT_IOReadINode():
*   Routine to read a INODE from the b+ tree file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successsful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOReadINode(BTREE *btree,
                   PTR offset, 
                   INODE *node
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fread(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
      return result;
    } else {
      printf("BT_IOReadINode : read node failure \n");
      return result;
    }
  } else {
    printf("BT_IOReadINode : seek node failure \n");
    return result;
  }
}

/***************************************************************************

* BT_IOWriteLNode():
*   Routine to write a LNODE to file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOWriteLNode(BTREE *btree, /* the b+ tree */ 
                    PTR offset,   /* the offset at which to write the node */
                    LNODE *node  /* the content of the node */
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fwrite(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
    }
  }

  return result;
}

/***************************************************************************

* BT_IOWriteONode():
*   Routine to write a ONODE to file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOWriteONode(BTREE *btree, /* the b+ tree */ 
                    PTR offset,   /* the offset at which to write the node */
                    ONODE *node  /* the content of the node */
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fwrite(node, O_PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
    }
  }

  return result;
}

/***************************************************************************

* BT_IOWriteINode():
*   Routine to write a INODE to file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_IOWriteINode(BTREE *btree, /* the b+ tree */ 
                    PTR offset,   /* the offset at which to write the node */
                    INODE *node  /* the content of the node */
) {
  int result = FALSE;

  IO_PageAccess(btree->BTreeFile, offset);

  if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
    if (fwrite(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      result = TRUE;
    }
  }

  return result;
}

/***************************************************************************

* BT_IOWriteNewLNode():
*   Routine to write a new LNODE to the file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   The offset of the new node in the file.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The node will be append at the end of the file
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
PTR BT_IOWriteNewLNode(BTREE *btree, /* the b+ tree */
                       LNODE *node   /* the content of the node */
) {
  PTR offset;

  if (fseek(btree->BTreeFile, 0L, SEEK_END)==0) {
    offset = ftell(btree->BTreeFile);
    IO_PageAccess(btree->BTreeFile, offset);
    if (fwrite(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      return offset;
    }
  }

  printf("BT_IOWriteNewLNode : error writing node to file \n");
  return (-1L);
}

/***************************************************************************

* BT_IOWriteNewONode():
*   Routine to write a new ONODE to the file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   The offset of the new node in the file.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The node will be append at the end of the file
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
PTR BT_IOWriteNewONode(BTREE *btree, /* the b+ tree */
                       ONODE *node   /* the content of the node */
) {
  PTR offset;
  
  if ((btree->top) > -1) {
     offset = btree->Stacks[btree->top];
     (btree->top)--;
     if (fseek(btree->BTreeFile, offset, SEEK_SET)==0) {
        IO_PageAccess(btree->BTreeFile, offset);
        if (fwrite(node, O_PAGE_SIZE, 1, btree->BTreeFile)) {
           return offset;
        }     
     }   
  } else {
    if (fseek(btree->BTreeFile, 0L, SEEK_END)==0) {
       offset = ftell(btree->BTreeFile);
       IO_PageAccess(btree->BTreeFile, offset);
       if (fwrite(node, O_PAGE_SIZE, 1, btree->BTreeFile)) {          
          return offset;
       }
    }
  }
  printf("BT_IOWriteNewONode : error writing node to file \n");
  return (-1L);
}

/***************************************************************************

* BT_IOWriteNewINode():
*   Routine to write a new INODE to the file
*
* IMPORTS:
*   None.
*
* RETURNS:
*   The offset of the new node in the file.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The node will be append at the end of the file
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
PTR BT_IOWriteNewINode(BTREE *btree, /* the b+ tree */
                       INODE *node   /* the content of the node */
) {
  PTR offset;

  if (fseek(btree->BTreeFile, 0L, SEEK_END)==0) {
    offset = ftell(btree->BTreeFile);
    IO_PageAccess(btree->BTreeFile, offset);
    if (fwrite(node, PAGE_SIZE, 1, btree->BTreeFile)) {
      return offset;
    }
  }

  printf("BT_IOWriteNewINode : error writing node to file \n");
  return (-1L);
}
