/*
 * Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, School of Computing, National University of Singapore, 2005.
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The library will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - No redistribution of the code is allowed
 *  - All bug fixes will be returned to zhangru1@comp.nus.edu.sg for
 *    inclusion in future releases
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <float.h>

#include "btree.h"
#include "nulllist.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define K 10
#define NOQUERY 200

extern int BT_Insert(BTREE *, const long, const KEYTYPE, RECORD_TYPE *);
extern int BT_Open(BTREE *, FILE *, PTR, PTR);
extern int BT_Close(BTREE *);

extern int BT_NewLNode(LNODE **);
extern int BT_IOReadLNode(BTREE *, PTR, LNODE *);
extern void IO_ResetPageAccess(void);

extern int BT_NewONode(ONODE **);
extern int BT_IOReadONode(BTREE *, PTR, ONODE *);

extern IMINODE* LoadINodes(PTR, BTREE *);
extern PTR IMBT_Locate(IMINODE *, KEYTYPE);
extern IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key);

extern int QueueSize;
extern long PageAccess; 
int got=0;
int max_top=-1;
extern void compact(BTREE *btree);

FILE *fp_result;
BTREE tree; 

PTR candidate[K];
DATA_TYPE distance[K];
int NoAnswer, FarthestNo;
KEYTYPE Farthest;
RECORD_TYPE query;
int curpar;
DATA_TYPE min_r[ClusterNumber], max_r[ClusterNumber];
KEYTYPE keylow[ClusterNumber], keyhigh[ClusterNumber];
PTR pagelow[ClusterNumber], pagehigh[ClusterNumber], donelow[ClusterNumber+1], donehigh[ClusterNumber+1];

void ProcessPoint(LENTRY le);
void SearchLeftNoCheck(PTR address, float curkeylow);
void SearchLeft(PTR address, float curkeylow);
void SearchRight(PTR address, float curkeyhigh);

main(int argc, char ** argv) {   
	
	float ClusterPoint[ClusterNumber][D];   
	char string[20];
	
	register float temp;
	
	DATA_TYPE dis_value;
	DATA_TYPE dis_list[ClusterNumber], Mindist;
	int MindistNo, count;
	register int i, j, p, k,x,y;
	long offset=0L;
	PTR address;
	RECORD_TYPE record;      
	int total, totalCan;
	FILE *fp_data;
	FILE *fp_query;
	FILE *fp_tree;   
	FILE *fp_btreeinfo;
	FILE *fp_reference;
	
	long pointer,over;
	int ContinueFlag,Deadline;
	long leafNumber=0L;
	long overNumber=0L; 
	long entryNumber=0L;
	long o_entryNumber=0L;
	DATA_TYPE leafUsage;
	
	DATA_TYPE IO[NOQUERY];
	DATA_TYPE averageIO=0.0;  
	INODE *iNode;  
	LNODE *lNode; 
	ONODE *oNode;  


	DATA_TYPE rad[K];
	
	int new_intersect[ClusterNumber];
	PTR Start[ClusterNumber];
	PTR Prev[ClusterNumber];
	PTR Next[ClusterNumber];
	
	DATA_TYPE R[ClusterNumber];
	DATA_TYPE R1[ClusterNumber];
	DATA_TYPE R2[ClusterNumber];
	DATA_TYPE r1[ClusterNumber], r2[ClusterNumber]; 
	DATA_TYPE s1_radius[ClusterNumber], s2_radius[ClusterNumber];
	DATA_TYPE go_on, searched;
	DATA_TYPE r, dd, DD;
	DATA_TYPE TotalMaxDist=0;
	
	KEYTYPE curkeylow, curkeyhigh, radius, r0, dr;
	
	IMINODE *root;
	
	clock_t t1,t2;
	time_t time1,time2;
	
	/*cluster centers*/
	fp_reference=fopen("reference","r");
	fseek(fp_reference,0L,SEEK_SET);
	for (i=0; i<ClusterNumber; i++) {
		for (j=0;j<D;j++) {
			fscanf(fp_reference,"%f",&(ClusterPoint[i][j]));
		}
	}
	fclose(fp_reference); 
	
	/*show reference points
	int myi,myj;
	for(myi=0;myi<ClusterNumber;myi++)
	{
		printf("Cluster %d\n",myi);
		for(myj=0;myj<D;myj++)
		{
			printf("%f ",ClusterPoint[myi][myj]);
			//printf("\n");
		}
		printf("\n");
	}
	//End show reference points*/  
	
	/*build indexing tree*/ 
	if(argc==2 && strcmp (argv [1], "b") == 0)
	{
		for (i=0; i<ClusterNumber; i++) {
			min_r[i]=FLT_MAX;   
			max_r[i]=0.0;
		}
		
		QueueSize=0;
		IO_ResetPageAccess();  
		max_top=-1;
		fp_tree=fopen("index","w+");  
		BT_Open(&tree,fp_tree,-1,-1);  

		/////define data file here///////
		fp_data=fopen("data","r"); 

		fseek(fp_data,0L,SEEK_SET);
		
		while(fread(&record,sizeof(RECORD_TYPE),1,fp_data)==1) {
			dis_value=idistance(record,ClusterPoint);
			PageAccess=0L;
			BT_Insert(&tree,offset,(KEYTYPE)dis_value,&record);
			offset++;
			
			i = (int)(dis_value/CST);

			if (dis_value > max_r[i]) max_r[i]=dis_value; //max_r[i] stores the max key in partition i
			
			if (dis_value < min_r[i]) min_r[i]=dis_value; //min_r[i] stores the min key in partition i
		}
		
		printf("\nThe tree is set up.\n");  
		printf("Root: %d\n",tree.Root);
		printf("Left most leaf node: %d \n", tree.Leaf_Head);
		printf("Max_top of Stack : %d \n", max_top);  
		
		
		/*back up btreeinfo information*/
		fp_btreeinfo=fopen("btreeinfo","w+");

		fprintf (fp_btreeinfo, "%ld\n", tree.Root);
		fprintf (fp_btreeinfo, "%ld\n", tree.Leaf_Head);

		for (i=0;i<ClusterNumber;i++) {
			fprintf(fp_btreeinfo,"%f ",min_r[i]);
			fprintf(fp_btreeinfo,"%f\n",max_r[i]);   
		}
		printf("\n btreeinfo done\n");  
		
		fclose(fp_btreeinfo);
		fclose(fp_data);		
		
		/*adding Prev point into leaf nodes*/
		
		fp_tree=fopen("index","r+");
		compact(&tree); 
		refine(&tree);
		BT_Close(&tree);
	}
	
	/*k-N-N*/
	else{
	
		t1=clock();
		time1=time(NULL);	
		
		fp_result=fopen("result","w+");
		fp_btreeinfo=fopen("btreeinfo","r"); 
		fscanf (fp_btreeinfo, "%ld", &(tree.Root));
      	fscanf (fp_btreeinfo, "%ld", &(tree.Leaf_Head));		
		for (i=0;i<ClusterNumber;i++) {
			fscanf(fp_btreeinfo,"%f",&min_r[i]);
			fscanf(fp_btreeinfo,"%f",&max_r[i]);        
		}
		fclose(fp_btreeinfo);
		IO_ResetPageAccess();
		fp_tree=fopen("index","r");
		BT_Open(&tree,fp_tree,tree.Root,tree.Leaf_Head);
		
		BT_NewLNode(&lNode);   
		BT_NewONode(&oNode);		
		
		/////////define query file here/////////////
		fp_query=fopen("sample","r");//the query file
		
		fseek(fp_query,0L,SEEK_SET);
		count=0;
		
		root=LoadINodes(tree.Root, &tree);
		
		while(fread(&query,sizeof(RECORD_TYPE),1,fp_query) && count<NOQUERY){
			
			PageAccess=0L;
			NoAnswer=0;
			Farthest=0;
			FarthestNo=0;			
			for(i=0;i<ClusterNumber;i++) keylow[i]=keyhigh[i]=donelow[i]=donehigh[i]=-1;
			donelow[ClusterNumber]=donehigh[ClusterNumber]=-2;
			
			//calculate distance between query and each reference point, and find the min distance
			for(i=MindistNo=0, Mindist=FLT_MAX; i<ClusterNumber; i++) 
			{
				dd=0.0;
				for (j=0;j<D;j++) {
					temp=query.fields[j]-ClusterPoint[i][j];
					dd+=temp*temp;
				}
				dis_list[i] = pow(dd,0.5) + i*CST;
				
				if(dd<Mindist)
				{
					Mindist=dd;
					MindistNo=i;
				}
			}
			
			address = IMBT_Locate(root,dis_list[MindistNo]);
			BT_IOReadLNode(&tree,address,lNode);
			
			for(i=0, Mindist=FLT_MAX;i<lNode->NumEntry;i++)
			{
				for(j=dd=0;j<D;j++)
				{
					temp = lNode->Entries[i].ver[j]-query.fields[j];
					dd += temp*temp;
				}
				if(Mindist>dd && dd!=0) Mindist=dd;
			}
			
			r0=radius=0.5*pow(Mindist,0.5);
			//r0=radius=0.05f;
			
			//dr=0.1*r0;
			dr=0.005f;
			
			//printf("r0:%f dr:%f ", r0, dr);

			NoAnswer=0;
			Farthest=0;
			FarthestNo=0;
			
			ContinueFlag=1;
			
			while(ContinueFlag)
			{
				curpar=0;
				if(dis_list[curpar]-curpar*CST <= radius+max_r[curpar])
				{
					curkeylow=dis_list[curpar]-radius;
					if(curkeylow < min_r[curpar]) curkeylow=min_r[curpar];
					
					curkeyhigh=dis_list[curpar]+radius;
					if(curkeyhigh > max_r[curpar]) curkeyhigh=max_r[curpar];
					
					if (keylow[curpar]<0) //this partition is not searched before
					{
						address=IMBT_Locate(root,dis_list[curpar]);
						BT_IOReadLNode(&tree,address,lNode);						
						for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i]);
						
						over=lNode->Overflow;				
						while(over >= 0)
						{
							BT_IOReadONode(&tree,over,oNode);
							for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i]);
							over=oNode->Next;
						}
						
						if(lNode->Prev<0) 
						{
							keylow[curpar]=min_r[curpar];
							donelow[curpar]=address;
						}
						else if(lNode->Entries[0].Key >= curkeylow)
							SearchLeftNoCheck(lNode->Prev, curkeylow);
						else
						{
							keylow[curpar]=lNode->Entries[0].Key;
							pagelow[curpar]=lNode->Prev;
							donelow[curpar]=address;
						}
						
						if(lNode->Next<0) 
						{
							keyhigh[curpar]=max_r[curpar];
							donehigh[curpar]=address;
						}
						else if(lNode->Entries[lNode->NumEntry-1].Key <= curkeyhigh)
							SearchRight(lNode->Next, curkeyhigh);
						else
						{
							keyhigh[curpar]=lNode->Entries[lNode->NumEntry-1].Key;
							pagehigh[curpar]=lNode->Next;
							donehigh[curpar]=address;
						}
					}
					
					else 
					{
						if(keylow[curpar]>curkeylow)
							SearchLeftNoCheck(pagelow[curpar], curkeylow);
						
						if(pagehigh[curpar]==donelow[curpar+1]) keyhigh[curpar]=max_r[curpar];
						else if(keyhigh[curpar]<curkeyhigh)
							SearchRight(pagehigh[curpar], curkeyhigh);
					}
				}
							
				for(curpar=1;curpar<ClusterNumber;curpar++)
				{
					if(dis_list[curpar]-curpar*CST <= radius+max_r[curpar])
					{
						curkeylow=dis_list[curpar]-radius;
						if(curkeylow < min_r[curpar]) curkeylow=min_r[curpar];
						
						curkeyhigh=dis_list[curpar]+radius;
						if(curkeyhigh > max_r[curpar]) curkeyhigh=max_r[curpar];
						
						if (keylow[curpar]<0) //this partition is not searched before
						{
							address=IMBT_Locate(root,dis_list[curpar]);
							BT_IOReadLNode(&tree,address,lNode);						
							for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i]);
							
							over=lNode->Overflow;							
							while(over >= 0)
							{
								BT_IOReadONode(&tree,over,oNode);
								for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i]);
								over=oNode->Next;
							}
							
							if(lNode->Prev<0 || lNode->Prev ==donehigh[curpar-1]) 
							{
								keylow[curpar]=min_r[curpar];
								donelow[curpar]=address;
							}
							else if(lNode->Entries[0].Key >= curkeylow)
								SearchLeft(lNode->Prev, curkeylow);
							else
							{
								keylow[curpar]=lNode->Entries[0].Key;
								pagelow[curpar]=lNode->Prev;
								donelow[curpar]=address;
							}
							
							if(lNode->Next<0 || lNode->Next == donelow[curpar+1]) 
							{
								keyhigh[curpar]=max_r[curpar];
								donehigh[curpar]=address;
							}
							else if(lNode->Entries[lNode->NumEntry-1].Key <= curkeyhigh)
								SearchRight(lNode->Next, curkeyhigh);
							else
							{
								keyhigh[curpar]=lNode->Entries[lNode->NumEntry-1].Key;
								pagehigh[curpar]=lNode->Next;
								donehigh[curpar]=address;
							}
						}
						
						else
						{
							if(pagelow[curpar]==donehigh[curpar-1]) keylow[curpar]=min_r[curpar];
							else if(keylow[curpar]>curkeylow)
								SearchLeft(pagelow[curpar], curkeylow);
							
							if(pagehigh[curpar]==donelow[curpar+1]) keyhigh[curpar]=max_r[curpar];
							else if(keyhigh[curpar]<curkeyhigh)
								SearchRight(pagehigh[curpar], curkeyhigh);
						}
					}
				}
				if(NoAnswer==K && Farthest<=radius*radius) ContinueFlag=0;				
				else radius += dr;
			}

#define SORTFLAG
#ifdef SORTFLAG
			float minDist,tempDist;
			long tempOffset;
			int minNo;
			
			for(i=0;i<K-1;i++)
			{
				minDist=distance[i];
				minNo=i;
				for(j=i+1;j<K;j++)
				{
					if(distance[j]<minDist)
					{
						minDist=distance[j];
						minNo=j;
					}
				}
				
				tempDist=distance[minNo];
				tempOffset=candidate[minNo];
				distance[minNo]=distance[i];
				candidate[minNo]=candidate[i];
				distance[i]=tempDist;
				candidate[i]=tempOffset;
			}
#endif
			fprintf(fp_result,"QueryPoint %d\n", count);
			for (i=0; i<K; i++) {
				fprintf(fp_result,"%ld %f\n",candidate[i], pow(distance[i],0.5));
			}
			fprintf(fp_result,"KNNDist: %f" , pow(Farthest,0.5));
			//fprintf(fp_result," r0: %f; radius: %f",r0, radius);
			fprintf(fp_result,"\n");
			fprintf(fp_result,"PageAccess:%ld\n",PageAccess);
			fprintf(fp_result,"\n");
			
			TotalMaxDist+=pow(Farthest,0.5);
			IO[count]=(DATA_TYPE)PageAccess;
			count++;
		}
		
		free(lNode);
		free(oNode);
		
		fprintf(fp_result,"%d queries\n",count);
		
		for(i=1;i<count;i++) {
			IO[0]+=IO[i];
		}
		averageIO=(DATA_TYPE)(IO[0])/count;
		fprintf(fp_result,"Average Page-IO: %f\nAverage KNN distance: %f\n",averageIO,TotalMaxDist/count);
		t2=clock();
		time2=time(NULL);
		
		fprintf(fp_result,"\nTotal clock:%ld\n",t2-t1);
		fprintf(fp_result,"Total time: %f\n",difftime(time2,time1));
		fprintf(fp_result,"Average query processing time: %f second\n", (float)(difftime(time2,time1))/count);
		
		BT_Close(&tree);
		fclose(fp_query);   
		fclose(fp_result);
		printf("\nKNN search is done. See the file \"result\" for the result.\n\n");      
	}
}

#define CHECKSAME

void ProcessPoint(LENTRY le)
{
	register int i, j;
	KEYTYPE CurDist;
	register KEYTYPE temp;
	
	for(i=CurDist=0;i<D;i++) 
	{
		temp = le.ver[i]-query.fields[i];
		CurDist += temp*temp;
	}
	
	if(NoAnswer<K)
	{
		
#ifdef CHECKSAME
		for (i=0;i<NoAnswer && candidate[i]!=le.Offset ;i++);
		if(i<NoAnswer) return;
#endif		
		candidate[NoAnswer]=le.Offset;
		distance[NoAnswer]=CurDist;
		if(CurDist>Farthest)
		{
			Farthest=CurDist;
			FarthestNo=NoAnswer;
		}
		NoAnswer++;
	}
	
	else
	{
		if(CurDist<Farthest)
		{
#ifdef CHECKSAME
			for (i=0;i<NoAnswer && candidate[i]!=le.Offset ;i++);
			if(i<NoAnswer) return;
#endif
			candidate[FarthestNo]=le.Offset;
			distance[FarthestNo]=CurDist;
		}
	
		//find the farthest point and record it
		Farthest=0;
		for(j=0;j<K;j++)
		{
			if(distance[j]>Farthest)
			{
				Farthest=distance[j];
				FarthestNo=j;
			}
		}
		//End find the farthest point and record it
	}	
}


void SearchLeftNoCheck(PTR address, float curkeylow)
{
	int i, j;
	int continueflag;
	ONODE *oNode;
	LNODE *lNode;
	PTR over;

	if(address<0) return;

	BT_NewLNode(&lNode);   
	BT_NewONode(&oNode);
	
	continueflag=1;
	
	while(continueflag)
	{
		BT_IOReadLNode(&tree,address,lNode);
		
		for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i]);
		
		over=lNode->Overflow;
		
		while(over >= 0)
		{
			BT_IOReadONode(&tree,over,oNode);
			for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i]);
			over=oNode->Next;
		}
		
		if(lNode->Entries[0].Key < curkeylow || lNode->Prev<0)
		{
			continueflag=0;
			donelow[curpar]=address;
		}		
		else address = lNode->Prev;
	}
	
	keylow[curpar]=lNode->Entries[0].Key;
	pagelow[curpar]=lNode->Prev;
	
	free(lNode);
	free(oNode);
}


void SearchLeft(PTR address, float curkeylow)
{
	int i, j;
	int continueflag;
	ONODE *oNode;
	LNODE *lNode;
	PTR over;

	if(address<0) return;

	BT_NewLNode(&lNode);   
	BT_NewONode(&oNode);
	
	continueflag=1;
	
	while(continueflag>0)
	{
		BT_IOReadLNode(&tree,address,lNode);
		
		for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i]);
		
		over=lNode->Overflow;
		
		while(over >= 0)
		{
			BT_IOReadONode(&tree,over,oNode);
			for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i]);
			over=oNode->Next;
		}
		
		if(lNode->Entries[0].Key < curkeylow || lNode->Prev<0)
		{
			continueflag=0;
			donelow[curpar]=address;
		}
		else if(lNode->Prev==donehigh[curpar-1])
		{
			continueflag=-1;
			keylow[curpar]=min_r[curpar];
		}
		else address = lNode->Prev;
	}
	
	if(continueflag==0)
	{
		keylow[curpar]=lNode->Entries[0].Key;
		pagelow[curpar]=lNode->Prev;
	}
	free(lNode);
	free(oNode);
}

void SearchRight(PTR address, float curkeyhigh)
{
	int i, j;
	int continueflag;
	ONODE *oNode;
	LNODE *lNode;
	PTR over;
	
	if(address<0) return;

	BT_NewLNode(&lNode);   
	BT_NewONode(&oNode);
	
	continueflag=1;
	
	while(continueflag>0)
	{
		BT_IOReadLNode(&tree,address,lNode);
		
		for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i]);
		
		over=lNode->Overflow;
		
		while(over >= 0)
		{
			BT_IOReadONode(&tree,over,oNode);
			for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i]);
			over=oNode->Next;
		}
		
		if(	lNode->Entries[lNode->NumEntry-1].Key > curkeyhigh || lNode->Next<0) 
		{
			continueflag=0;
			donehigh[curpar]=address;
		}
		else if(lNode->Next==donelow[curpar+1])
		{
			continueflag=-1;
			keyhigh[curpar]=max_r[curpar];
		}
		else address = lNode->Next;
	}
	
	if(continueflag==0)
	{
		keyhigh[curpar]=lNode->Entries[lNode->NumEntry-1].Key;
		pagehigh[curpar]=lNode->Next;
	}
	free(lNode);
	free(oNode);	
}

