/*
 * Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, School of Computing, National University of Singapore, 2005.
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The library will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - No redistribution of the code is allowed
 *  - All bug fixes will be returned to zhangru1@comp.nus.edu.sg for
 *    inclusion in future releases
 *
 */

/*************************************************************************

*
* File : nulllist.c
*
* Modules:
* Routines to manipulate nulllist nodes
*   int NL_NodeFull()
*   void NL_AddToNode()
*   int NL_NewNode()
*   void InitNode()
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/
#include "nulllist.h"
#include "btree.h"


/* function prototype */
/* internal */

int NL_NodeFull(const NL_NODE *);
void NL_AddToNode(NL_NODE *, const long);
int NL_NewNode(NL_NODE **);
void NL_InitNode(NL_NODE *);
void NL_PrintNode(NL_NODE *);

/***************************************************************************

* NL_NodeFull():
*   Routine to check whether a NL_Node is full
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if the node is full else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int NL_NodeFull(const NL_NODE *node  /* the node to br checked */
) {
  if ((node->Header).NumEntry == NODE_FAN_OUT){
   return TRUE;
  } else {
   return FALSE;
  }
}

/***************************************************************************

* NL_AddToNode():
*   Routine to add an entry to a node
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void NL_AddToNode(NL_NODE *node,
                  const long recid
) {
  (node->Offsets)[(node->Header).NumEntry++] = recid;
}

/***************************************************************************

* functionName:
*   ShortDescription
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int NL_NewNode(NL_NODE **newnode
) {
  if (*newnode = (NL_NODE *)malloc(sizeof(NL_NODE))){
    return TRUE;
  }else{
    return FALSE;
  }
}

/***************************************************************************

* NL_InitNode():
*   Routine to inintialise an empty node
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   If a non-empty node is passed in, content of the node will be overwritten.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void NL_InitNode(NL_NODE *node    /* the node to br initialised */
) {
  (node->Header).NumEntry = 0;
}

/***************************************************************************

* NL_PrintNode():
*   Routine to print the content of a node for debugging purpose
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void NL_PrintNode(NL_NODE *node
) {
  int num;
  int i;

  num = (node->Header).NumEntry;
  printf("The number of entry in the node is : %d\n", num);
  for (i=0; i<num; i++){
    printf("Entry %d : %d\n", i, node->Offsets[i]);
  }
}
