/*
 * Version 1.0 (Beta Test)
 * Copyright (c) Rui Zhang, School of Computing, National University of Singapore, 2005.
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The library will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - No redistribution of the code is allowed
 *  - All bug fixes will be returned to zhangru1@comp.nus.edu.sg for
 *    inclusion in future releases
 *
 */

/*************************************************************************

*
* File : nulllist_io.c
*
* Modules:
* Routines to perform IO functions for nulllist nodes
*   int NL_IOReadNode()
*   int NL_IOWriteNode()
*   int NL_IOWriteNewNode()
*
* Version 1
* Coded by: Dai Haoyu		Modified by:
* Date: 12/9/97			Date modified: 
*************************************************************************/
#include <assert.h>
#include "nulllist.h"
#include "btree.h"


/* function prototype */
/* internal */
int NL_IOReadNode(FILE *, const long, NL_NODE *);
int NL_IOWriteNode(FILE *, const long, NL_NODE *);
int NL_IOWriteNewNode(FILE *, NL_NODE *);

/* external */
extern void IO_PageAccess(FILE *, long);

/***************************************************************************

* NL_IOReadNode():
*   Routine to read a nullist node from disk to memory
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if reading is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int NL_IOReadNode(FILE *nlfp, /* the index from which to 
                                             * read the node        */
                  const long offset,        /* the offset of the node in
                                             * the index file       */
                  NL_NODE *node             /* the address where the node
                                             * will be read to      */
) {
  int result = FALSE;

  IO_PageAccess(nlfp, offset);

  if (fseek(nlfp, offset*PAGE_SIZE, SEEK_SET)==0){
    if (fread(node, PAGE_SIZE, 1, nlfp)){
      result = TRUE;
      return result;
    } else {
      printf("NL_IOReadNode : read node failure \n");
      return result;
    }
  } else {
    printf("NL_IOReadNode : seek node failure \n");
    return result;
  }
}

/***************************************************************************

* NL_IOWriteNode():
*   Routine to write a node back to disk
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if write is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int NL_IOWriteNode(FILE *nlfp,   /* the index to which to
                                                * to write the node   */
                   const long offset,          /* the offset of the node */
                   NL_NODE *node               /* the node to be written */
) {
  int result = FALSE;

  IO_PageAccess(nlfp, offset);

  if (fseek(nlfp, offset*PAGE_SIZE, SEEK_SET)==0){
    if (fwrite(node, PAGE_SIZE, 1, nlfp)){
      result = TRUE;
    }
  }
  return result;
}

/***************************************************************************

* NL_IOWriteNewNode():
*   Routine to write a new node to disk
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if write is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   It will append the node at the end of the file.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int NL_IOWriteNewNode(FILE *nlfp,  /* the index to which to write 
                                                  * the node            */
                      NL_NODE *node              /* the node to be written */
) {
  int result = FALSE;
  long offset;

  if (fseek(nlfp, 0L, SEEK_END)==0){
    offset = ftell(nlfp) / PAGE_SIZE;
    IO_PageAccess(nlfp, offset);
    if (fwrite(node, PAGE_SIZE, 1, nlfp)){
      result = TRUE;
    }
  }

  return result;
}
